/*
 * Copyright 2012 Hanson Robokind LLC.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.robokind.api.common.osgi;

import java.util.List;
import java.util.Map;
import java.util.Properties;

/**
 *
 * @author Matthew Stevenson <www.robokind.org>
 */
public abstract class AbstractLifecycleProvider<I,T extends I> implements 
        ServiceLifecycleProvider<I> {

    private List<ServiceRequirementDescriptor> myDescriptors;
    protected T myService;
    protected Properties myRegistrationProperties;
    
    public AbstractLifecycleProvider(List<ServiceRequirementDescriptor> reqs){
        if(reqs == null){
            throw new NullPointerException();
        }
        myDescriptors = reqs;
    }

    @Override
    public void start(Map<String, Object> dependencies) {
        if(dependencies == null){
            throw new NullPointerException();
        }
        if(!Validator.validateServices(myDescriptors, dependencies)){
            throw new IllegalArgumentException(
                    "Invalid Dependency set for factory.");
        }
        myService = create(dependencies);
    }
    
    protected abstract T create(Map<String,Object> services);

    @Override
    public T getService() {
        return myService;
    }
    
    @Override
    public Properties getRegistrationProperties(){
        return myRegistrationProperties;
    }

    @Override
    public List<ServiceRequirementDescriptor> getServiceRequirements() {
        return myDescriptors;
    }

    @Override
    public void requiredServiceChanged(String serviceId, Object service) {
        if(serviceId == null){
            throw new NullPointerException();
        }
        validate(serviceId, service);
        handleChange(serviceId, service);
    }
    
    protected abstract void handleChange(String serviceId, Object service);
    
    private void validate(String id, Object req){
        List<ServiceRequirementDescriptor> reqs = getServiceRequirements();
        if(req != null){
            if(Validator.validateService(reqs, id, req)){
                return;
            }
            throw new IllegalArgumentException("Invalid service or id.  "
                    + "id: " + id + ", service: " + req);
        }
        if(Validator.validateServiceId(reqs, id)){
            return;
        }
        throw new IllegalArgumentException("Invalid service id: " + id + ".");
    }
    
}
