/*
 * Copyright 2012 Hanson Robokind LLC.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.robokind.api.common.osgi;

import org.robokind.api.common.osgi.ServiceLifecycleProvider.Validator;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceRegistration;
import static org.robokind.api.common.osgi.ServiceRequirementsTracker.*;

/**
 *
 * @author Matthew Stevenson <www.robokind.org>
 */
public class DynamicServiceLauncher<T>{
    private final static Logger theLogger = 
            Logger.getLogger(DynamicServiceLauncher.class.getName());
    private BundleContext myContext;
    private ServiceLifecycleProvider<T> myLifecycleProvider;
    private ServiceRequirementsTracker myRequirementsTracker;
    private T myService;
    private ServiceRegistration myServiceRegistration;
    private Properties myRegistrationProperties;
    private String myRegistrationClassName;
    private boolean myInitializedFlag;
    
    public DynamicServiceLauncher(
            BundleContext context, ServiceLifecycleProvider<T> factory){
        this(context, factory, factory.getServiceClass().getName(), null);
    }
    
    public DynamicServiceLauncher(
            BundleContext context, ServiceLifecycleProvider<T> factory, 
            Properties registrationProps){
        this(context, factory, 
                factory.getServiceClass().getName(), registrationProps);
    }
    
    public DynamicServiceLauncher(
            BundleContext context, ServiceLifecycleProvider<T> provider, 
            String registrationClassName, Properties registrationProps){
        if(context == null || provider == null 
                || registrationClassName == null){
            throw new NullPointerException();
        }
        myContext = context;
        myLifecycleProvider = provider;
        List<ServiceRequirementDescriptor> descriptors = 
                myLifecycleProvider.getServiceRequirements();
        if(descriptors == null){
            throw new NullPointerException();
        }
        myRequirementsTracker = new ServiceRequirementsTracker(myContext);
        for(ServiceRequirementDescriptor dd : descriptors){
            myRequirementsTracker.addRequirement(
                    dd.getServiceClass(), 
                    dd.getRequirementId(), 
                    dd.getServiceFilter());
        }
        myRequirementsTracker.addPropertyChangeListener(
                new RequirementStatusListener());
        myRegistrationClassName = registrationClassName;
        myRegistrationProperties = registrationProps;
        myInitializedFlag = false;
    }
    
    public void start(){
        myRequirementsTracker.start();
    }
    
    public void stop(){
        myRequirementsTracker.stop();
    }
    
    public void unregisterService(){
        myService = null;
        if(myServiceRegistration == null){
            return;
        }
        myServiceRegistration.unregister();
        myServiceRegistration = null;
    }
    
    private void handleRequiredServices(Map<String,Object> requiredServices){
        if(myService != null){
            return;
        }else if(!ServiceLifecycleProvider.Validator.validateServices(
                myLifecycleProvider.getServiceRequirements(), 
                requiredServices)){
            throw new IllegalArgumentException(
                    "Invalid requirement set for service.");
        }
        myLifecycleProvider.start(requiredServices);
        myService = myLifecycleProvider.getService();
        if(myService == null){
            theLogger.warning("The factory failed to create a service "
                    + "and returned null.");
            return;
        }
        theLogger.log(Level.INFO, 
                "Service created of type {0}", 
                myLifecycleProvider.getServiceClass());
        registerService();
    }
    
    private void handleChanged(String id, Object newService){
        if(id == null){
            throw new NullPointerException();
        }else if(!validate(id, newService)){
            throw new IllegalArgumentException("Invalid service or id.  "
                    + "id: " + id + ", required service: " + newService);
        }
        myLifecycleProvider.requiredServiceChanged(id, newService);
        checkForModification();
    }
    
    private boolean validate(String id, Object req){
        List<ServiceRequirementDescriptor> reqs = 
                myLifecycleProvider.getServiceRequirements();
        return (req == null && Validator.validateServiceId(reqs, id)) ||
                Validator.validateService(reqs, id, req);
    }
    
    private void checkForModification(){
        T service = myLifecycleProvider.getService();
        if(service == null && myService != null){
            theLogger.warning("Required Service change stopped this service.  "
                    + "Service is being unregistered.");
            unregisterService();
        }else if(service != null && myService == null){
            theLogger.info(
                    "Required Service change has started this service.");
            myService = service;
            registerService();
        }else if(service != myService){
            theLogger.warning(
                    "Required Service change has changed this service.");
            ServiceRegistration oldReg = myServiceRegistration;
            myService = service;
            registerService();
            oldReg.unregister();
        }
    }
    
    protected void registerService(){
        Properties props = myLifecycleProvider.getRegistrationProperties();
        Properties allProps = null;
        if(props != null || myRegistrationProperties != null){
            allProps = new Properties();
            if(props != null){
                allProps.putAll(props);
            }
            if(myRegistrationProperties != null){
                allProps.putAll(myRegistrationProperties);
            }
        }
        ServiceRegistration reg = myContext.registerService(
                myRegistrationClassName, myService, allProps);
        myServiceRegistration = reg;
        theLogger.log(Level.INFO, 
                "Service Successfully Registered: {0}", myService.toString());
    }
    
    class RequirementStatusListener implements PropertyChangeListener{

        @Override
        public void propertyChange(PropertyChangeEvent evt) {
            if(evt == null){
                return;
            }
            String name = evt.getPropertyName();
            if(PROP_ALL_REQUIREMENTS_AVAILABLE.equals(name)){
                handleAllAvailable(evt);
            }else if(!myInitializedFlag){
                return;
            }else if(PROP_REQUIREMENT_CHANGED.equals(name)){
                String id = evt.getOldValue().toString();
                Object obj = evt.getNewValue();
                handleChanged(id, obj);
            }
        }
        
        private boolean handleAllAvailable(PropertyChangeEvent evt){
            theLogger.info("All requirements available, "
                    + "attempting to create service.");
            Object obj = evt.getNewValue();
            if(obj == null || !(obj instanceof Map)){
                theLogger.warning(
                        "Invalid requirement map, cannot create service.");
                return true;
            }try{
                Map<String,Object> reqs = (Map<String,Object>)obj;
                handleRequiredServices(reqs);
            }catch(ClassCastException ex){
                theLogger.log(Level.WARNING, 
                        "Improper requirement Map type.", ex);
                return true;
            }
            myInitializedFlag = true;
            return true;
        }
    }
}
