/*
 * Copyright 2012 Hanson Robokind LLC.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.robokind.api.common.osgi;

import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Logger;

/**
 *
 * @author Matthew Stevenson <www.robokind.org>
 */
public interface ServiceLifecycleProvider<T> {
    /**
     * Called the first time all requirements are available.
     * @param requiredServices
     */
    public void start(Map<String,Object> requiredServices);
    /**
     * Called after the service is started and a requirement is changed.
     * @param serviceId
     * @param service 
     */
    public void requiredServiceChanged(
            String serviceId,  Object service);
    /**
     * Returns the service provided.  Returns null if not started or if
     * requirements have been removed and it is no longer available.
     * @return 
     */
    public T getService();
    
    /**
     * Returns a list describing the services required to start this service.
     * @return 
     */
    public List<ServiceRequirementDescriptor> getServiceRequirements();
    
    public Properties getRegistrationProperties();
    
    public Class<T> getServiceClass();
    
    public static class Validator{
        private final static Logger theLogger = 
                Logger.getLogger(Validator.class.getName());
        public static boolean validateServices(
                List<ServiceRequirementDescriptor> descriptors, 
                Map<String,Object> services){
            for(ServiceRequirementDescriptor descriptor : descriptors){
                if(!checkDescriptor(descriptor, services)){
                    return false;
                }
            }
            return true;
        }

        private static boolean checkDescriptor(
                ServiceRequirementDescriptor descriptor, 
                Map<String,Object> requirements){
            String id = descriptor.getRequirementId();
            Object req = requirements.get(id);
            if(req == null){
                return false;
            }
            Class reqClass = descriptor.getServiceClass();
            if(!reqClass.isAssignableFrom(req.getClass())){
                return false;
            }
            return true;
        }
        
        public static boolean validateService(
                List<ServiceRequirementDescriptor> descriptors,
                String requirmentId, Object requirement){
            if(requirmentId == null || requirement == null){
                theLogger.warning("Found null argument.  Returning false.");
                return false;
            }
            for(ServiceRequirementDescriptor descriptor : descriptors){
                String id = descriptor.getRequirementId();
                Class reqClass = descriptor.getServiceClass();
                if(requirmentId.equals(id) &&
                        reqClass.isAssignableFrom(requirement.getClass())){
                    return true;
                }
            }
            return false;
        }
        
        public static boolean validateServiceId(
                List<ServiceRequirementDescriptor> descriptors,
                String requirementId){
            if(requirementId == null){
                theLogger.warning("Found null argument.  Returning false.");
                return false;
            }
            for(ServiceRequirementDescriptor descriptor : descriptors){
                String id = descriptor.getRequirementId();
                if(requirementId.equals(id)){
                    return true;
                }
            }
            return false;
        }
    }
}
