/*
 * Copyright 2012 Hanson Robokind LLC.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.robokind.api.common.osgi;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.osgi.framework.BundleContext;
import org.robokind.api.common.property.PropertyChangeNotifier;

/**
 * Tracks a number of Services required for our Service.
 * When all requirements are available, our Service is created and registered.
 * If a requirement becomes unavailable, our Service is unregistered until the
 * requirements is available.
 * 
 * @author Matthew Stevenson <www.robokind.org>
 */
public class ServiceRequirementsTracker extends PropertyChangeNotifier{
    private final static Logger theLogger = 
            Logger.getLogger(ServiceRequirementsTracker.class.getName());
    public final static String PROP_REQUIREMENT_AVAILABLE = "requirementAvailable";
    public final static String PROP_REQUIREMENT_CHANGED = "requirementChanged";
    public final static String PROP_REQUIREMENT_UNAVAILABLE = "requirementUnavailable";
    public final static String PROP_ALL_REQUIREMENTS_AVAILABLE = "allRequirementsAvailable";
    
    private BundleContext myContext;
    private List<SingleServiceListener> myRequirementTrackers;
    private Map<String,SingleServiceListener> myRequirementIds;
    private Map<String,Object> myAvailableRequirements;
    private boolean myListeningFlag;
    
    public ServiceRequirementsTracker(BundleContext context){
        if(context == null){
            throw new NullPointerException();
        }
        myContext = context;
        myRequirementTrackers = new ArrayList<SingleServiceListener>();
        myAvailableRequirements = new HashMap();
        myRequirementIds = new HashMap<String, SingleServiceListener>();
        myListeningFlag = false;
    }
    
    public boolean requirementsSatisfied(){
        return myAvailableRequirements.size() == myRequirementTrackers.size();
    }
    /**
     * Returns a map of requirementIds to requirements.  Returns null until all
     * requirements are available.
     * @return map of requirementIds to requirements, null if all
     * requirements are not available
     */
    public Map<String,Object> getRequiredServices(){
        if(!requirementsSatisfied()){
            return null;
        }
        return myAvailableRequirements;
    }
    
    public Object getRequiredService(String requirementId){
        return myAvailableRequirements.get(requirementId);
    }
    
    public void addRequirement(
            Class clazz, String requirementId, String filterString){
        if(clazz == null || requirementId == null){
            throw new NullPointerException();
        }
        if(myListeningFlag){
            throw new IllegalStateException(
                    "Unable to add requirement while listening.");
        }else if(myRequirementIds.containsKey(requirementId)){
            throw new IllegalArgumentException(
                    "Unable to add requirement, requirementId already exists.");
        }
        SingleServiceListener ssl = 
                new SingleServiceListener(clazz, myContext, filterString);
        ssl.addPropertyChangeListener(new RequirementListener(requirementId));
        myRequirementTrackers.add(ssl);
        myRequirementIds.put(requirementId, ssl);
    }
    
    public void start(){
        myListeningFlag = true;
        for(SingleServiceListener ssl : myRequirementTrackers){
            ssl.start();
        }
    }
    
    public void stop(){
        myListeningFlag = false;
        myAvailableRequirements.clear();
        for(SingleServiceListener ssl : myRequirementTrackers){
            ssl.stop();
        }
    }
    
    private void requiredServiceFound(String requirementId, Object req){
        if(requirementId == null || req == null){
            throw new NullPointerException();
        }
        if(!myRequirementIds.containsKey(requirementId)){
            return;
        }
        theLogger.log(Level.INFO, "Found required service: {0}", requirementId);
        myAvailableRequirements.put(requirementId, req);
        firePropertyChange(PROP_REQUIREMENT_AVAILABLE, requirementId, req);
        checkAllRequirements();
    }
    
    private void checkAllRequirements(){
        if(!requirementsSatisfied()){
            return;
        }
        theLogger.log(Level.INFO, "All requirements present: {0}", 
                Arrays.toString(myRequirementIds.keySet().toArray()));
        firePropertyChange(PROP_ALL_REQUIREMENTS_AVAILABLE, null, getRequiredServices());
        
    }
    
    private void requiredServiceChanged(String requirementId, Object req){
        if(requirementId == null || req == null){
            throw new NullPointerException();
        }
        if(!myRequirementIds.containsKey(requirementId)){
            return;
        }
        theLogger.log(Level.INFO, "Required service changed: {0}", requirementId);
        myAvailableRequirements.put(requirementId, req);
        firePropertyChange(PROP_REQUIREMENT_CHANGED, requirementId, req);
    }
    
    private void requiredServiceLost(String requirementId){
        if(requirementId == null){
            throw new NullPointerException();
        }
        if(!myRequirementIds.containsKey(requirementId)){
            return;
        }
        theLogger.log(Level.INFO, "Lost required service: {0}", requirementId);
        myAvailableRequirements.remove(requirementId);
        firePropertyChange(PROP_REQUIREMENT_UNAVAILABLE, requirementId, null);
    }
    
    class RequirementListener implements PropertyChangeListener {
        private String myRequirementId;
        
        public RequirementListener(String requirementId){
            myRequirementId = requirementId;
        }
        
        @Override
        public void propertyChange(PropertyChangeEvent evt) {
            if(SingleServiceListener.PROP_SERVICE_TRACKED.equals(evt.getPropertyName())){
                track(evt.getOldValue(), evt.getNewValue());
            }else if(SingleServiceListener.PROP_SERVICE_REMOVED.equals(evt.getPropertyName())){
                untrack(evt.getNewValue());
            }
        }
        
        private void track(Object oldVal, Object newVal){
            if(oldVal == null){
                requiredServiceFound(myRequirementId, newVal);
            }else{
                requiredServiceChanged(myRequirementId, newVal);
            }
        }
        
        private void untrack(Object obj){
            requiredServiceLost(myRequirementId);
        }
    }
}
